import click

from examtool.api.database import get_exam
from examtool.api.extract_questions import extract_questions
from examtool.api.scramble import scramble
from examtool.api.substitutions import get_all_substitutions
from examtool.cli.deploy import get_name
from examtool.cli.utils import exam_name_option


@click.command()
@exam_name_option
@click.option("--email", help="The target student's email address.")
@click.option("--show-all", default=False, help="Show all questions received, not just with substitutions.")
def substitutions(exam, email, show_all):
    """
    Show the substitutions a particular student received
    """
    original_exam = get_exam(exam=exam)
    exam = get_exam(exam=exam)
    exam = scramble(email, exam, keep_data=True)
    question_substitutions = get_all_substitutions(original_exam, exam)
    questions = extract_questions(exam)
    for question in questions:
        substitutions = question_substitutions[question["id"]]
        if substitutions or show_all:
            print(get_name(question), substitutions)



import json
import os
import pathlib

import click

from examtool.api.database import get_roster, get_logs
from examtool.cli.utils import exam_name_option, hidden_output_folder_option


@click.command()
@exam_name_option
@hidden_output_folder_option
def save_logs(exam, out):
    """
    Save the full submission log for later analysis.
    Note that this command is slow.
    To view a single log entry, run `examtool log`.
    """
    out = out or "out/logs/" + exam

    pathlib.Path(out).mkdir(parents=True, exist_ok=True)

    roster = get_roster(exam=exam)
    for i, (email, deadline) in enumerate(roster):
        print(email)
        logs = get_logs(exam=exam, email=email)
        with open(os.path.join(out, email), "w") as f:
            json.dump(logs, f)



import json

from examtool.api.database import get_exam, get_roster
from examtool.api.extract_questions import extract_questions
from examtool.api.scramble import scramble
from google.cloud import firestore
import warnings
warnings.filterwarnings("ignore", "Your application has authenticated using end user credentials")


db = firestore.Client()
exams = [x.id for x in db.collection("exams").stream()]

for exam in exams:
    print("checking", exam)
    exam_json = json.dumps(get_exam(exam=exam))
    roster = get_roster(exam=exam)

    flagged = set()

    for email, _ in roster:
        template_questions = extract_questions(json.loads(exam_json))
        student_questions = list(
            extract_questions(scramble(email, json.loads(exam_json), keep_data=True))
        )
        student_question_lookup = {q['id']: q for q in student_questions}
        for question in template_questions:
            if question["id"] not in student_question_lookup:
                continue
            if question["type"] not in ["multiple_choice", "select_all"]:
                continue
            if question["id"] in flagged:
                continue

            for i, option in enumerate(question["options"]):
                option["index"] = i

            s = lambda options: sorted(options, key=lambda q: q["text"])

            for a, b in zip(s(question["options"]), s(student_question_lookup[question["id"]]["options"])):
                if a["index"] != b.get("index", a["index"]):
                    flagged.add(question["id"])
                    continue

    if flagged:
        print(exam, flagged)



import json

from examtool.api.database import get_exam
from examtool.api.extract_questions import extract_questions
from examtool.api.scramble import scramble, get_elements


def find_unexpected_words(exam, logs):
    data = get_exam(exam=exam)
    exam_json = json.dumps(data)
    for i, (email, log) in enumerate(logs):
        all_alternatives = get_substitutions(data)
        selected = {
            question["id"]: question["substitutions"]
            for question in extract_questions(scramble(email, json.loads(exam_json), keep_data=True))
        }
        for record in log:
            record.pop("timestamp")
            question = next(iter(record.keys()))
            answer = next(iter(record.values()))
            if question not in all_alternatives:
                continue
            for keyword, variants in all_alternatives[question].items():
                for variant in variants:
                    if variant == selected[question][keyword]:
                        continue
                    if variant in answer:
                        # check for false positive
                        for other in selected[question].values():
                            if variant in other and other != variant:
                                break
                        else:
                            print(email, selected[question], variant, answer)
                            break
                else:
                    continue
                break
            else:
                continue
            break


def get_substitutions(exam):
    out = {}

    def process_group(group, substitutions):
        group_substitutions = group["substitutions"]
        for element in get_elements(group):
            if element.get("type") == "group":
                process_group(element, {**substitutions, **group_substitutions})
            else:
                process_question(element, {**substitutions, **group_substitutions})

    def process_question(question, substitutions):
        out[question["id"]] = {**substitutions, **question["substitutions"]}

    global_substitutions = exam["substitutions"]
    for group in exam["groups"]:
        process_group(group, global_substitutions)

    return out



import csv
from datetime import datetime

import click
import pytz

from examtool.api.database import get_logs
from examtool.cli.utils import exam_name_option


def time(timestamp):
    return datetime.utcfromtimestamp(timestamp).replace(tzinfo=pytz.utc).astimezone(pytz.timezone("America/Los_Angeles"))


@click.command()
@click.option("--email", default=None, help="A student's target email address.")
@click.option("--roster", default=None, type=click.File("r"), help="A roster CSV.")
@exam_name_option
def logs(email, exam, roster):
    """
    Export the submission history.
    Specify `email` to target a particular student,
    or specify `roster` to target all students listed in a particular roster CSV.
    """
    if roster:
        roster = csv.reader(roster, delimiter=",")
        next(roster)
        emails = [x[0] for x in roster if x[1]]
    else:
        emails = [email]

    all_students = []

    for email in emails:
        print(email)
        times = []
        for record in get_logs(exam=exam, email=email):
            ref = time(record.pop("timestamp"))
            times.append([ref, next(iter(record.keys())), next(iter(record.values()))])
        print("\n".join(str(x) + " " + str(y) + " " + str(z) for x, y, z in sorted(times)))

    all_students.sort(key=lambda x: x[0])

    print("\n".join(str(x) + " " + str(y) for x, y in all_students))


if __name__ == '__main__':
    logs()



import click

from examtool.cli.utils import exam_name_option, hidden_output_folder_option
import examtool.api.download


@click.command()
@exam_name_option
@click.option(
    "--name-question",
    default=None,
    help="The ID of the question for the student's name.",
)
@click.option(
    "--sid-question", default=None, help="The ID of the question for the student's SID."
)
@click.option(
    "--compact/--expanded",
    default=False,
    help="Combine questions on the same page to save space.",
)
@click.option(
    "--with-substitutions/--without-substitutions",
    default=False,
    help="Include keyword substitutions in exported question bodies.",
)
@hidden_output_folder_option
def download(exam, out, name_question, sid_question, compact, with_substitutions):
    """
    Download student submissions for an exam.
    Exams are downloaded as PDFs into a target folder - specify `out` to redirect the folder.
    An `OUTLINE.pdf` is also generated for Gradescope, as is a `summary.csv` for analytics or autograding.
    """
    exam_json, template_questions, email_to_data_map, total = examtool.api.download.download(
        exam
    )
    examtool.api.download.export(
        template_questions,
        email_to_data_map,
        total,
        exam,
        out,
        name_question,
        sid_question,
        compact,
        substitute_in_question_text=with_substitutions,
    )


if __name__ == "__main__":
    download()



from examtool.api.scramble import is_compressible_group


def extract_questions(exam, extract_public_bool: bool = True, top_level: bool = True, include_groups: bool = False):
    def group_questions(group):
        out = _group_questions(group)
        try:
            first = next(out)
        except StopIteration:
            return
        first["text"] = group["name"] + "\n\n" + group["text"] + "\n\n" + first["text"]
        yield first
        yield from out

    def _group_questions(group):
        for i, element in enumerate(group.get("elements", []) + group.get("questions", [])):
            element["index"] = f"{group.get('index', '0.')}{i + 1}."
            if element.get("type") == "group":
                if include_groups:
                    yield element
                out = group_questions(element)
                yield from out
            else:
                yield element

    if extract_public_bool:
        yield from extract_public(exam)
    if top_level:
        for i, group in enumerate(exam["groups"]):
            group["index"] = str(i + 1) + "."
            if include_groups:
                yield group
            yield from group_questions(group)
    else:
        yield from group_questions(exam)


def extract_public(exam):
    if exam.get("public"):
        yield from extract_questions(exam["public"], extract_public_bool=False, top_level=False)


def extract_groups(group):
    for g in group["groups"]:
        if is_compressible_group(g):
            for g2 in g["elements"]:
                yield g2
        else:
            yield g



import re

import click

from datetime import date, datetime
from collections import OrderedDict

exam_name_option = click.option(
    "--exam", prompt=True, default="cs61a-test-final", help="The exam name."
)
hidden_output_folder_option = click.option(
    "--out",
    default=None,
    help="Output folder. Leave as default for dependent commands to work.",
    type=click.Path()
)
hidden_target_folder_option = click.option(
    "--target",
    default=None,
    help="Target folder for PDFs. Leave as default unless the source output folder is not the default.",
    type=click.Path()
)


def prettify(course_code):
    m = re.match(r"([a-z]+)([0-9]+[a-z]?)", course_code)
    return m and (m.group(1) + " " + m.group(2)).upper()

def determine_semester():
    sem_map = OrderedDict()
    sem_map["1/1"] = "Spring"
    sem_map["6/1"] = "Summer"
    sem_map["9/1"] = "Fall"
    today = date.today()
    def compare_date(DATE_TO_COMPARE):
        parsed = datetime.strptime(DATE_TO_COMPARE, '%m/%d').date().replace(year=today.year)
        return parsed > today
    sem = next(iter(sem_map))
    for DATE_TO_COMPARE, new_sem in sem_map.items():
        if compare_date(DATE_TO_COMPARE):
            break
        sem = new_sem
    return f"{sem} {today.year}"



import json
import os

import click

from examtool.api.database import get_roster
from examtool.api.substitution_finder import find_unexpected_words
from examtool.cli.utils import exam_name_option, hidden_target_folder_option


@click.command()
@exam_name_option
@hidden_target_folder_option
def cheaters(exam, target):
    """
    Identify potential instances of cheating.
    """
    if not target:
        target = "out/logs/" + exam
    logs = []
    for email, deadline in get_roster(exam=exam):
        with open(os.path.join(target, email)) as f:
            logs.append([email, json.load(f)])
    find_unexpected_words(exam, logs)



from functools import wraps


def as_list(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        return list(func(*args, **kwargs))
    return wrapped


def sanitize_email(email):
    return email.replace("_", r"\_")





import json
import os
import pathlib
from datetime import datetime
from io import BytesIO

from pikepdf import Pdf, Encryption
import click
import pytz

from examtool.api.database import get_exam, get_roster
from examtool.api.utils import sanitize_email
from examtool.api.scramble import scramble
from examtool.api.gen_latex import render_latex
from examtool.cli.utils import determine_semester, exam_name_option, hidden_output_folder_option, prettify


@click.command()
@exam_name_option
@click.option("--subtitle", prompt=True, default="Structure and Interpretation of Computer Programs")
@hidden_output_folder_option
@click.option("--do-twice", is_flag=True, help="Run the compile twice for each student to fix weird rendering bugs.")
@click.option("--email", help="The email address of a particular student.")
@click.option(
    "--exam-type",
    default="Final Exam",
    help="The type of exam you are given. For example 'Final Exam' (default).",
)
@click.option(
    "--semester",
    default=determine_semester(),
    help=f"The semester of the exam. '{determine_semester()}' (default).",
)
@click.option("--deadline", default=None, help="Generates exam regardless of if student is in roster with the set deadline.")
def compile_all(exam, subtitle, out, do_twice, email, exam_type, semester, deadline):
    """
    Compile individualized PDFs for the specified exam.
    Exam must have been deployed first.
    """
    if not out:
        out = "out/latex/" + exam

    pathlib.Path(out).mkdir(parents=True, exist_ok=True)

    exam_data = get_exam(exam=exam)
    password = exam_data.pop("secret")[:-1]
    print(password)
    exam_str = json.dumps(exam_data)
    
    roster = get_roster(exam=exam)

    if email:
        roster = [line_info for line_info in roster if line_info[0] == email]
        if len(roster) == 0:
            if deadline:
                roster = [(email, deadline)]
            else:
                raise ValueError("Email does not exist in the roster!")


    for email, deadline in roster:
        if not deadline:
            continue
        exam_data = json.loads(exam_str)
        scramble(email, exam_data)
        deadline_utc = datetime.utcfromtimestamp(int(deadline))
        deadline_pst = pytz.utc.localize(deadline_utc).astimezone(
            pytz.timezone("America/Los_Angeles")
        )
        deadline_string = deadline_pst.strftime("%I:%M%p")

        with render_latex(
            exam_data,
            {
                "emailaddress": sanitize_email(email),
                "deadline": deadline_string,
                "coursecode": prettify(exam.split("-")[0]),
                "description": subtitle,
                "examtype": exam_type,
                "semester": semester,
            },
            do_twice=do_twice,
        ) as pdf:
            pdf = Pdf.open(BytesIO(pdf))
            pdf.save(
                os.path.join(
                    out, "exam_" + email.replace("@", "_").replace(".", "_") + ".pdf"
                ),
                encryption=Encryption(owner=password, user=password),
            )
            pdf.close()


if __name__ == "__main__":
    compile_all()



import os

from examtool.api.server_delegate import server_only

from sendgrid import SendGridAPIClient


@server_only
def send_email(*, exam, data):
    sg = SendGridAPIClient(os.environ.get("SENDGRID_API_KEY"))
    sg.client.mail.send.post(data)


@server_only
def send_email_batch(*, exam, data_list):
    sg = SendGridAPIClient(os.environ.get("SENDGRID_API_KEY"))
    for data in data_list:
        sg.client.mail.send.post(data)


@server_only
def get_api_key(*, exam):
    return os.environ.get("SENDGRID_API_KEY")


def send_email_local(key, data):
    sg = SendGridAPIClient(key)
    response = sg.client.mail.send.post(data)
    print(response.status_code)
    print(response.body)
    print(response.headers)

def send_email_local_safe(key, data, timeout=10):
    """
    Sends an email, and raises a BadResponse if either the response code is not
        2xx or it takes more than `timeout` seconds to respond
    """
    from func_timeout import func_timeout, FunctionTimedOut
    try:
        def do_send_email():
            sg = SendGridAPIClient(key)
            response = sg.client.mail.send.post(data)
            if not 200 <= response.status_code < 300:
                raise BadResponse("{}\n{}\n{}".format(response.status_code, response.body, response.headers))
        func_timeout(timeout, do_send_email)
    except FunctionTimedOut:
        raise BadResponse("Timed out!")

class BadResponse(Exception):
    pass



import random


def scramble(email, exam, *, keep_data=False):
    random.seed(email)

    def scramble_group(group, substitutions, config, depth):
        group_substitutions = select(group["substitutions"])
        group_substitutions.update(
            select_no_replace(group.get("substitutions_match", []))
        )
        substitute(
            group,
            [*substitutions, group_substitutions],
            ["name", "html", "tex", "text"],
        )
        if depth in config["scramble_groups"] or group.get("scramble"):
            scramble_keep_fixed(get_elements(group))
        if group.get("pick_some"):
            get_elements(group)[:] = random.sample(
                get_elements(group), group["pick_some"]
            )
        for element in get_elements(group):
            if element.get("type") == "group":
                scramble_group(
                    element, [*substitutions, group_substitutions], config, depth + 1
                )
            else:
                scramble_question(
                    element, [*substitutions, group_substitutions], config
                )

        if is_compressible_group(group):
            text, html, tex = group["text"], group["html"], group["tex"]
            element = get_elements(group)[0]
            if element.get("type") != "group" and depth == 1:
                return
            group.clear()
            group.update(element)
            group["text"] = text + "\n" + group["text"]
            group["html"] = html + "\n" + group["html"]
            group["tex"] = tex + "\n" + group["tex"]

    def scramble_question(question, substitutions, config):
        question_substitutions = select(question["substitutions"])
        question_substitutions.update(
            select_no_replace(question.get("substitutions_match", []))
        )
        substitute(
            question, [question_substitutions, *substitutions], ["html", "tex", "text"]
        )
        if isinstance(question["options"], list):
            if "scramble_options" in config:
                scramble_keep_fixed(question["options"])
            for option in question["options"]:
                substitute(
                    option,
                    [*substitutions, question_substitutions],
                    ["html", "tex", "text"],
                )

        if keep_data and "solution" in question:
            solution = question["solution"]
            if solution.get("solution") is not None:
                substitute(
                    solution["solution"],
                    [question_substitutions, *substitutions],
                    ["html", "tex", "text"],
                    store=False,
                )
            else:
                options = solution["options"]
                if options:
                    substitute(
                        options,
                        [question_substitutions, *substitutions],
                        range(len(options)),
                        store=False,
                    )
        else:
            question.pop("solution", None)

    def substitute(target: dict, list_substitutions, attrs, *, store=True):
        merged = {}
        for substitutions in list_substitutions:
            merged = {**merged, **substitutions}
            for attr in attrs:
                for k, v in substitutions.items():
                    target[attr] = target[attr].replace(k, v)
                    target[attr] = target[attr].replace(k.title(), v.title())
                    target[attr] = target[attr].replace(
                        latex_escape(k), latex_escape(v)
                    )
        if store:
            if keep_data:
                target["substitutions"] = merged
            else:
                target.pop("substitutions", None)

    def scramble_keep_fixed(objects):
        if keep_data:
            for i, object in enumerate(objects):
                object["index"] = i
        movable_object_pos = []
        movable_object_values = []
        for i, object in enumerate(objects):
            if not object.get("fixed"):
                movable_object_pos.append(i)
                movable_object_values.append(object)
        random.shuffle(movable_object_values)
        for i, object in zip(movable_object_pos, movable_object_values):
            objects[i] = object

    global_substitutions = select(exam["substitutions"])
    global_substitutions.update(select_no_replace(exam.get("substitutions_match", [])))
    exam["config"]["scramble_groups"] = exam["config"].get(
        "scramble_groups", [-1]
    ) or range(100)
    if 0 in exam["config"]["scramble_groups"]:
        scramble_keep_fixed(exam["groups"])
    for group in exam["groups"]:
        scramble_group(group, [global_substitutions], exam["config"], 1)
    exam.pop("config", None)

    return exam


def get_elements(group):
    return group.get("elements") if "elements" in group else group.get("questions")


def select(substitutions):
    out = {}
    # DEFINE
    for k, v in sorted(substitutions.items()):
        out[k] = random.choice(v)
    return out


def select_no_replace(substitutions_match):
    out = {}
    # DEFINE MATCH
    for item in substitutions_match:
        k = item["directives"]
        v = item["replacements"]
        values = v.copy()
        for choice in k:
            c = random.choice(values)
            values.remove(c)
            out[choice] = c
    return out


def is_compressible_group(group):
    return (
        group.get("pick_some") == 1
        and not group["name"].strip()
        and group["points"] is None
    )


def latex_escape(text):
    return text.replace("_", r"\_")



import os
from collections import defaultdict

import click

from examtool.cli.utils import exam_name_option


@click.command()
@exam_name_option
def check_dupes(exam):
    """
    Search for PDFs submitted for multiple exams.
    If any are found, you must manually check that the correct one is uploaded to Gradescope.
    """
    files = defaultdict(list)
    for target in os.listdir("out/export"):
        if not target.startswith(exam):
            continue
        target = os.path.join("out/export", target)
        if not os.path.isdir(target):
            continue
        for file in os.listdir(target):
            if "@" not in file:
                continue
            files[file].append(target)
    for file, exams in files.items():
        if len(exams) > 1:
            print(file, exams)


if __name__ == '__main__':
    check_dupes()

